# {NotifyHub} is a callback facility. {NotifyHub} is used by the informer
# to notify clients about arbitrary events in the informer. {NotifyHub}
# contains notification sets ({NotifySet}).
#
# {NotifySet} is identified by set ID and it can include one or many
# notifications ({Notify}). Client defines the action performed at
# notification (callback). Informer activates the notification when
# notification events occur.
#
# Notifications can be enabled/disabled on different levels
# ({NotifyHub}, {NotifySet}, and {Notify}).
#
# Usage example:
#
#    require 'notifyhub'
#
#    # Create class that includes interesting events.
#    class Storage
#
#        # Handle to NotifyHub.
#        attr_accessor :hub
#
#        def initialize
#            # Create NotifyHub with 3 callbacks.
#            @hub = NotifyHub.declare( :store, :load, :na )
#        end
#
#        # Store data and notify clients.
#        def store( data )
#            @data = data
#            @hub[ :store ].notify( @data )
#        end
#
#        # Load data and notify clients.
#        def load
#            @hub[ :load ].notify( @data )
#            @data
#        end
#    end
#
#
#    storage = Storage.new
#
#    # Setup notify action for store.
#    storage.hub[ :store ].action do |data|
#        puts "store: #{data}"
#    end
#
#    # Setup notify action for load.
#    storage.hub[ :load ].action do |data|
#        puts "load:  #{data}"
#    end
#
#
#    # Use storage and get notifications.
#    storage.store( "my data" )
#    data = storage.load
#
#
# Produces:
#   store: my data
#   load:  my data
class NotifyHub


    # Declare sets automatically at action registration.
    attr_accessor :autodeclare


    # Non-existing {Notify} error.
    class NotFound < RuntimeError; end

    # Redefining {Notify} error.
    class Redefining < RuntimeError; end


    # Create {NotifyHub} and declare sets.
    #
    # @param id [Array<Symbol>] Notify set(s).
    def NotifyHub.declare( *id )
        NotifyHub.new( *id )
    end


    # Create {NotifyHub} and with autodeclare for sets. Declare sets
    # automatically at action registration.
    #
    # @param id [Array<Symbol>] Notify set(s).
    def NotifyHub.auto( *id )
        n = NotifyHub.new( *id )
        n.autodeclare = true
        n
    end


    # Create {NotifyHub} and run block with it.
    #
    # @yield Block to run with NotifyHub.
    def NotifyHub.create( &blk )
        cg = NotifyHub.new
        if block_given?
            cg.instance_eval( &blk )
        end
        cg
    end


    # Instantiation.
    #
    # @param id [Array<Symbol>] Notify set id(s).
    def initialize( *id )
        @autodeclare = false
        @set = {}
        declare( *id )
    end


    # Declare {Notify} by set. Multiple notifiers can exist per set.
    #
    # @param id [Array<Symbol>] Notify set id(s).
    def declare( *id )
        id.each do |i|
            if @set[ i ]
                raise Redefining, "Notify set already declared: #{i.to_s}"
            else
                @set[ i ] = NotifySet.new( i )
            end
        end
    end


    # Register action to {NotifySet}. Multiple notifies can
    # exist per set.
    #
    # @param id [Symbol] Notify set id (class).
    # @param action [Block] Notify action.
    def action( id, &action )
        useSet( id ) do |set|
            set.action( &action )
        end
    end

    alias register action
    alias with     action


    # Remove all or one Notify.
    #
    # @param id [Symbol] Notify set id (class).
    # @param notify [Notify] Notify to remove (all if not given).
    def remove( id, notify = nil )
        withSet( id ) do |set|
            set.remove( notify )
        end
    end


    # Enable/disable {Notify} set or all if not set.
    #
    # @param id [Symbol] Notify set id (all if nil given).
    # @param value [Boolean] Enable with true and disable with false.
    def enable( id, value )
        if id
            withSet( id ) do |set|
                set.enable = value
            end
        else
            @set.each_value do |set|
                set.enable = value
            end
        end
    end


    # Run all notifiers in {Notify} set.
    #
    # @param id [Symbol] Notify set id (class).
    # @param args [Array<Object>] Arguments for notifiers.
    def notify( id, *args )
        withSet( id ) do |set|
            set.notify( *args )
        end
    end


    # Get {NotifySet} by ID.
    #
    # @param id [Symbol] Set ID.
    # @return [NotifySet] Set.
    def []( id )
        useSet( id ) do |set|
            set
        end
    end


    # Get list of {NotifySet} IDs.
    #
    # @return [Array<Symbol>] NotifySet IDs.
    def ids
        @set.keys
    end



    private

    # Check that {NotifySet} exists (or create it).
    def withSet( id, &blk )
        if @set[id]
            yield @set[ id ]
        else
            raise NotFound, "Uknown notify set: #{id.to_s}"
        end
    end


    # Check that {NotifySet} exists (or create it).
    def useSet( id, &blk )

        if @set[id] || @autodeclare

            unless @set[id]
                declare( id )
            end

            yield @set[ id ]

        else

            raise NotFound, "Uknown notify set: #{id.to_s}"

        end
    end

end



# {Notify} object container that includes disable/enable features.
class NotifySet


    # Enable/disable {NotifySet}.
    attr_accessor :enable


    # Instantiation.
    #
    # @param id [Symbol] Id of the set.
    def initialize( id = nil )
        @id = id
        @enable = true
        @list = []
    end


    # Enable/disable {NotifySet}.
    #
    # @param value [Boolean] Enable with true and disable with false.
    def enable( value )
        @enable = value
    end


    # Register {Notify}.
    #
    # @param action [Block] {Notify} action.
    def action( &action )
        n = Notify.new( &action )
        @list.push n
        n
    end

    alias register action
    alias with     action


    # Remove all or one Notify.
    #
    # @param notify [Notify] {Notify} to remove (all if not given).
    def remove( notify = nil )
        if notify
            @list.delete( notify )
        else
            @list = []
        end
    end

    # Run all notifiers in {NotifySet}.
    #
    # @param args [Array<Object>] Arguments for notifiers.
    # @return [Array<Object>] Array or single callback return value.
    def notify( *args )
        ret = []
        if @enable
            @list.each do |n|
                ret.push n.notify( *args )
            end
        end

        if ret.length == 1
            # Return single value.
            ret[0]
        else
            # Return array of results.
            ret
        end
    end


    # Iterate over list of notifiers.
    def each
        @list.each do |n|
            yield n
        end
    end


    # Get {Notify} by index.
    def []( idx )
        @list[ idx ]
    end

end



# Notify object including disable/enable possibility.
class Notify

    # Enabled/disabled status.
    attr_reader :enable


    # Instantiation.
    #
    # @param action [Proc] Notification action.
    def initialize( &action )
        @action = action
        @enable = true
    end


    # Perform notification action if enabled.
    #
    # @param args [Array<Object>] Data provided to action.
    def notify( *args )
        if @enable
            @action.yield( *args )
        end
    end


    # Enable/disable {Notify}.
    #
    # @param value [Boolean] Enable with true and disable with false.
    def enable( value )
        @enable = value
    end

end


# Include version definitions.
require_relative 'version'
